/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is resource file for sqlite.
 *
 * The Initial Developer of the Original Code is Ryan VanderMeulen.
 * Portions created by the Initial Developer are Copyright (C) 2007
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include "sqlite3.h"
#include "sqlite-version.h"
#include <winver.h>

#define SQLITE_COMPANY_NAME "sqlite.org"
#define SQLITE_INTERNAL_NAME "sqlite3"
#define SQLITE_FILEDESCRIPTION "SQLite Database Library"

#define MY_FILEOS VOS_NT_WINDOWS32

/////////////////////////////////////////////////////////////////////////////
//
// Version-information resource
//

VS_VERSION_INFO VERSIONINFO

 FILEVERSION SQLITE_VERSION_MAJOR,SQLITE_VERSION_MINOR,SQLITE_VERSION_PATCH,SQLITE_VERSION_SUBPATCH
 PRODUCTVERSION SQLITE_VERSION_MAJOR,SQLITE_VERSION_MINOR,SQLITE_VERSION_PATCH,SQLITE_VERSION_SUBPATCH
 FILEOS MY_FILEOS
 FILETYPE VFT_DLL

BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040904B0" // Lang=US English, CharSet=Unicode
        BEGIN
            VALUE "CompanyName", SQLITE_COMPANY_NAME "\0"
            VALUE "FileDescription", SQLITE_FILEDESCRIPTION "\0"
            VALUE "FileVersion", SQLITE_VERSION "\0"
            VALUE "InternalName", SQLITE_INTERNAL_NAME "\0"
            VALUE "OriginalFilename", SQLITE_INTERNAL_NAME ".dll\0"
            VALUE "ProductName", SQLITE_FILEDESCRIPTION "\0"
            VALUE "ProductVersion", SQLITE_VERSION "\0"
        END
    END
    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x409, 1200
    END
END
