/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 * XPCTL : nsILE.h
 *
 * ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is mozilla.org code..
 *
 * The Initial Developer of the Original Code is
 * Sun Microsystems, Inc.
 * Portions created by the Initial Developer are Copyright (C) 2000
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Prabhat Hegde (prabhat.hegde@sun.com)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either of the GNU General Public License Version 2 or later (the "GPL"),
 * or the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#ifndef nsILE_h__
#define nsILE_h__

#include "nscore.h"
#include "nsISupports.h"
#include "nsCtlCIID.h"

/*
 * nsILE Interface declaration 
 */
class nsILE : public nsISupports {
public:

  NS_DECLARE_STATIC_IID_ACCESSOR(NS_ILE_IID)
  
  NS_IMETHOD NeedsCTLFix(const PRUnichar*, const PRInt32,
                         const PRInt32, PRBool *) = 0;

  NS_IMETHOD GetPresentationForm(const PRUnichar*, PRUint32,
                                 const char*, char*, PRSize*,
                                 PRBool = PR_FALSE) = 0;

  NS_IMETHOD PrevCluster(const PRUnichar*, PRUint32, 
                         const PRInt32, PRInt32*) = 0;

  NS_IMETHOD NextCluster(const PRUnichar*, PRUint32,
                         const PRInt32, PRInt32*) = 0;

  NS_IMETHOD GetRangeOfCluster(const PRUnichar*, PRUint32,
                               const PRInt32, PRInt32*, PRInt32*) = 0;
};

NS_DEFINE_STATIC_IID_ACCESSOR(nsILE, NS_ILE_IID)

#endif /* !nsILE_h */

